from enviroment import env, config
import os
import gtk
import gobject
import threading
import urllib2
import pycurl
import time
import groove
import re
import math

class Song(gobject.GObject):
    """
    This class encasuplates a handle song information.
    """
    
    """
    Song data example:
    {
        u'SongID': 8428135, 
        u'IsLowBitrateAvailable': 1, 
        u'AvgDuration': 199, 
        u'Year': '', 
        u'Popularity': 1128700160, 
        u'SongName': u'Me gustas mucho', 
        u'AlbumName': u'Especial', 
        u'TrackNum': 3, 
        u'AlbumID': 1318743, 
        u'ArtistID': 12057, 
        u'Score': 156174.67887245, 
        u'Flags': 0, 
        u'EstimateDuration': 198, 
        u'AvgRating': 0, 
        u'CoverArtFilename': u'1318743.jpg', 
        u'TSAdded': u'1210029617', 
        u'GenreID': 7, 
        u'IsVerified': 1, 
        u'ArtistName': u'Viejas Locas'
    }
    """
    STATE_NOT_STARTED = 0
    STATE_CONNECTING = 1
    STATE_DOWNLOADING = 2
    STATE_COMPLETED = 3
    STATE_PAUSED = 4
    STATE_CANCELED = 5
    STATE_ERROR = 6
    last_song_local_id = 0
    
    def __init__(self, data):
        """
        @param data: The array data given by GrooveShark. This array also can has a
        key called 'filename', it is used to initialize the Song with a filename,
        used mainly by the app when load the downqueue file to know on wich file should
        download the song.
        """
        self.__gobject_init__()
        Song.last_song_local_id += 1
        
        self.local_id = Song.last_song_local_id
        self.data = data
        self.cover_pixbuf = None
        self.download_progress = None
        self.file_size = None
        self.filename = None
        self.state = Song.STATE_NOT_STARTED
        self.download_thread = None
        self.last_error = None
        
        try:
            self.filename = self.data["filename"]
        except KeyError:
            self.filename = self.get_default_filename()
        
        if self.get_cover_filename() == "":
            self.set_cover_missed_pixbuf()
            
        self.connect("download-initializing", self.on_download_initializing)
        self.connect("download-started", self.on_download_started)
        self.connect("download-paused", self.on_download_paused)
        self.connect("download-canceled", self.on_download_canceled)
        self.connect("download-completed", self.on_download_completed)
        self.connect("download-error", self.on_download_error)
    
    def on_download_initializing(self, song):
        print "[Download connecting]", self.get_id()
        self.state = Song.STATE_CONNECTING
    
    def on_download_started(self, song):
        print "[Download started]", self.get_id()
        self.state = Song.STATE_DOWNLOADING
    
    def on_download_paused(self, song):
        print "[Download paused]", self.get_id()
        self.state = Song.STATE_PAUSED
        
    def on_download_canceled(self, song):
        print "[Download canceled]", self.get_id()
        self.state = Song.STATE_CANCELED
        
    def on_download_completed(self, song):
        print "[Download completed]", self.get_id()
        self.state = Song.STATE_COMPLETED
        
    def on_download_error(self, song, error):
        print "[Download error]", self.get_id(), error
        self.state = Song.STATE_ERROR
        self.last_error = error
    
    def __getitem__(self, key):
        """
        Allows to use the song as a dictionary (not recommended, use
        methods instead)
        """
        return self.data[key]
    
    def clone(self):
        """
        Clone the song but changes it local ID to avoid errors during the song searching
        on AbstractSongList
        """
        newsong = Song(self.get_data())
        newsong.set_cover_pixbuf(self.get_cover_pixbuf())
        newsong.set_filename(self.get_filename())
        
        return newsong
    
    def equals(self, othersong):
        """
        Compare a song with another
        @param othersong: the song to compare
        """
        if othersong == None:
            return False
        
        return self.local_id == othersong.local_id
    
    def get_data(self):
        return self.data
    
    def get_state(self):
        return self.state
    
    def get_id(self):
        return self.data['SongID']
        
    def get_title(self):
        return self.data['SongName']
    
    def get_album(self):
        return self.data['AlbumName']
        
    def get_artist(self):
        return self.data['ArtistName']
        
    def get_track_num(self):
        return self.data['TrackNum']
    
    def get_year(self):
        return self.data['Year']
    
    def get_duration(self):
        return self.data['EstimateDuration']
    
    @staticmethod
    def format_time(dur):
        duration = float(dur)
        mins = math.floor(duration / 60)
        secs = duration % 60
        
        return "%02d:%02d" % (mins, secs)
    
    def get_duration_human_readable(self):
        return Song.format_time(self.get_duration()) 
    
    def get_last_error(self):
        return self.last_error
    
    def get_cover_filename(self):
        return self.data['CoverArtFilename']
    
    def get_cover_pixbuf(self):
        """
        Get the pixbuf for the song album cover.
        If returns None, means that we need to download the cover
        using download_cover
        """
        return self.cover_pixbuf
        
    def set_cover_pixbuf(self, pixbuf):
        """
        Set the pixbuf for the song album cover.
        Used by SongCoverThread and clone.
        """
        self.cover_pixbuf = pixbuf
        
    def set_cover_missed_pixbuf(self):
        self.cover_pixbuf = gtk.gdk.pixbuf_new_from_file("%s/data/sdefault.png" % env().BASEPATH)
        
    def download_cover(self):
        """
        Download the album cover, the object will emit the 'cover-downloaded'
        signal when the download finish.
        """
        thread = SongCoverThread(self)
        thread.start()
        
    def start_download(self, restart = False, speed = None):
        """
        Starts the song download, the object will emit various signals to
        handle the download progress.
        @param restart: If is False means that we should try to continue with the
        download if file exists.
        @param speed: The speed limit for the download, the default is the confi-
        gured in the app.
        """
        if speed == None:
            speed = int(config()["speed_limit"])
        
        self.download_thread = DownloadThread(self)
        self.download_thread.restart = restart
        self.download_thread.speed = speed
        self.download_thread.start()
        
    def is_downloading(self):
        """
        Returns true if the download is running
        """
        return self.download_thread != None and self.download_thread.is_alive()
    
    def is_paused(self):
        """
        Returns true if the download was stopped by pausing it
        """
        return self.download_thread != None and self.is_downloading() == False and self.download_thread.is_canceled == False
        
    def pause_download(self):
        """
        Pause the current download
        """
        if self.is_downloading():
            self.download_thread.pause()
        
    def cancel_download(self):
        """
        Cancel the current download
        """
        if self.is_downloading():
            self.download_thread.cancel()
        else:
            if self.state != Song.STATE_COMPLETED:
                os.remove(self.get_filename())
            self.emit("download-canceled")
            
    def resume_download(self):
        """
        Resume the file download, the only difference with start_download is that
        this method checks wether the file is downloading, ignore the action
        in that case.
        """
        if self.is_downloading() == False and self.state != Song.STATE_COMPLETED:
            self.start_download(False)
        
    def pause_download_sync(self):
        """
        Pause the current download and wait until it stops altogether.
        """
        if self.is_downloading():
            self.download_thread.pause()
            self.download_thread.join()
            
    def cancel_download_sync(self):
        """
        Cancel the current download and wait until it stops altogether.
        """
        if self.is_downloading():
            self.download_thread.cancel()
            self.download_thread.join()
            
    def get_download_progress(self):
        """
        Returns the current download progress in bytes or None if
        the download is not initialized.
        """
        return self.download_progress

    def set_download_progress(self, value):
        self.download_progress = value

    def get_file_size(self):
        """
        Returns the current file size in bytes or None if
        the download is not initialized.
        """
        return self.file_size

    def set_file_size(self, value):
        self.file_size = value
    
    def get_filename(self):
        return self.filename
    
    def set_filename(self, value):
        self.filename = value
        
    def get_default_filename(self):
        pattern = config()['file_pattern'] + ".mp3"
        filename = pattern.format(artist = self.get_artist().strip("<>:\"/\|?&*"),
                                  song = self.get_title().strip("<>:\"/\|?&*"),
                                  album = self.get_album().strip("<>:\"/\|?&*"))
        filename = filename.replace('/', '-')
        filename = os.path.join(config()['down_path'], filename)
        
        return filename

gobject.type_register(Song)
"""
Signals of the Song type
"""
# Emitted when the cover download was finished
gobject.signal_new("cover-downloaded", Song, gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, ())
# Emitted when download starts or resumes
gobject.signal_new("download-initializing", Song, gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, ())
# Emitted when download starts or resumes
gobject.signal_new("download-started", Song, gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, ())
# Emitted when cUrl receive data from GrooveShark (useful for update download progress)
gobject.signal_new("download-updated", Song, gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, ())
# Emitted when the download is successfully downloaded
gobject.signal_new("download-completed", Song, gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, ())
# Emitted when download is canceled
gobject.signal_new("download-canceled", Song, gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, ())
# Emitted when download is paused
gobject.signal_new("download-paused", Song, gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, ())
# Emitted when download is stopped
gobject.signal_new("download-stopped", Song, gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, ())
# Emitted when a download error occurs, an error message is passed by parameter
gobject.signal_new("download-error", Song, gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, (gobject.TYPE_STRING,))


class SongCoverThread(threading.Thread):
    """
    This class download a song cover album and notify
    it when download ends. Should not be used directly, use Song instead.
    """
    
    def __init__(self, song):
        threading.Thread.__init__(self)
        self.song = song

    def run(self):
        cover_url = self.song.get_cover_filename()

        try:
            url = "http://beta.grooveshark.com/static/amazonart/s%s"
            response = urllib2.urlopen(url % cover_url)
            loader = gtk.gdk.PixbufLoader()
            loader.write(response.read())
            loader.close()
            self.song.set_cover_pixbuf(loader.get_pixbuf())
        except Exception, e:
            print "Error while downloading cover: ", e
            self.song.set_cover_missed_pixbuf()

        self.song.emit("cover-downloaded")
        

class DownloadThread(threading.Thread):
    """
    This class download the song file and notify the progress.
    Should not be used directly, use Song instead.
    """
    
    def __init__(self, song):
        threading.Thread.__init__(self)

        self.song = song
        self.restart = False
        self.speed = 0
        self.first_hook = True
        self._stop = threading.Event()
        self.is_canceled = True
        self.file = None
        self.curl = None

    def run(self):
        self.song.emit("download-initializing")
        
        self.first_hook = True
        self._stop.clear()
        restart = self.restart
        
        if restart == False and os.path.exists(self.song.get_filename()) == False:
            restart = True
        
        try:
            if restart == True:
                self.file = open(self.song.get_filename(), "wb")
            else:
                self.file = open(self.song.get_filename(), "ab")
        except IOError:
            self.song.emit("download-error", _("Failed to create '%s' for writing.") % self.song.get_filename())
            return
        
        try:
            key = groove.getStreamKeyFromSongIDEx(self.song.get_id())
        except Exception as e:
            self.song.emit("download-error", e.__str__())
            return

        try:
            url = "http://" + key["result"]["%s" % self.song.get_id()]["ip"] + "/stream.php"
            url = str(url)
            c = pycurl.Curl()
            c.setopt(pycurl.URL, url)
            c.setopt(pycurl.NOPROGRESS, 0)
            c.setopt(pycurl.PROGRESSFUNCTION, self.hook)
            c.setopt(pycurl.HEADERFUNCTION, self.header_hook)
            c.setopt(pycurl.FAILONERROR, True)
            c.setopt(pycurl.WRITEDATA, self.file)
            if restart == False:
                c.setopt(pycurl.RESUME_FROM, os.path.getsize(self.song.get_filename()))
            c.setopt(pycurl.POST, True)
            c.setopt(pycurl.NOSIGNAL, True)
            if self.speed != 0:
                c.setopt(pycurl.MAX_RECV_SPEED_LARGE, self.speed)
            c.setopt(pycurl.POSTFIELDS, str("streamKey=" + key["result"]["%s" % self.song.get_id()]["streamKey"]))
            
            c.perform()
            
            self.file.close()
            self.song.set_download_progress(100)
            self.song.emit("download-completed")
        except pycurl.error, e:
            self.file.close()
            if e[0] == pycurl.E_ABORTED_BY_CALLBACK:
                if self.is_canceled == True:
                    os.remove(self.song.get_filename())
                    self.song.emit("download-canceled")
                else:
                    self.song.emit("download-paused")
            else:
                os.remove(self.song.get_filename())
                self.song.emit("download-error", e[1])
        except Exception, e:
            self.file.close()
            os.remove(self.song.get_filename())
            self.song.emit("download-error", e.__str__())
            return
            
        self.song.emit("download-stopped")
    
    def cancel(self):
        self.is_canceled = True
        self._stop.set()
        
    def pause(self):
        self.is_canceled = False
        self._stop.set()
    
    def stopped(self):
        return self._stop.isSet()
    
    def header_hook(self, data):
        data = data.split("\r\n")
        matches = re.match(r"HTTP/\d\.\d (\d+)", data[0])
        if matches:
            status = matches.group(1)
            
            # If is not partial content (partial content is status code 206)
            if status == "200":
                self.file.truncate(0)
    
    def hook(self, downloadTotal, downloadCurrent, uploadTotal, uploadCurrent):
        if self.stopped():
            return True
        
        if downloadTotal > 0:
            if self.first_hook:
                self.first_hook = False
                self.song.emit("download-started")
            
            progress = (downloadCurrent / downloadTotal) * 100
            self.song.set_download_progress(progress)
            self.song.set_file_size(downloadTotal)

            if progress <= 99:
                self.song.emit("download-updated")
        
        return False